//================================================================================================
// $Id: ParserHelper.h 6309 2023-01-16 20:19:42Z aron $
//
// Project:
//		Welltype Parser
// Contents:
//		Parser helper
// Authors:
//		Aron Barath, 2014-2023
//================================================================================================
// Copyright (c) Aron Barath
//================================================================================================

#ifndef WTPARSER_PARSERHELPER_H
#define WTPARSER_PARSERHELPER_H

class WTFlexLexer;

#include <set>
#include <map>
#include <vector>
#include "frontend.h"
#include "problemhandler.h"
#include "sourcemanager.h"
#include "ast/ast_program.h"
#include "ast/ast_expression.h"
#include "ast/ast_type.h"
#include "ast/ast_typeinfo.h"
#include "ast/ast_functionsignature.h"
#include "ast/ast_functionargument.h"
#include "ast/ast_functioncall.h"
#include "ast/ast_operatordeclaration.h"
#include "ast/ast_variableaccess.h"
#include "debugger.h"
#include "symtab.h"
#include "uniqueid.h"

namespace wtparser
{

class WTParser;

class ParserHelper
{
public:
	ParserHelper(wtparser::ISourceManager* ism, wtparser::IProblemHandler* iph, const wtparser::IOptions* opt);
	~ParserHelper();

	bool Init(const std::string & sourcename)
	{
		return wtparser::SourceManager::OK == mSrcMan.OpenFile(sourcename);
	}

	std::string GetFilename() { return mSrcMan.GetFilename(true); }
	unsigned int GetLine();
	WTFlexLexer* GetLexer() { return mSrcMan.GetTopLexer(); }
	wtparser::ProblemHandler & GetProblemHandler() { return mProbHan; }
	wtparser::SourceManager & GetSourceManager() { return mSrcMan; }
	void PostError(const std::string & msg) { PostError(msg, GetLine()); }
	void PostError(const std::string & msg, unsigned int line);
	void PostWarning(const std::string & msg) { PostWarning(msg, GetLine()); }
	void PostWarning(const std::string & msg, unsigned int line);
	void PostNote(const std::string & msg) { PostNote(msg, GetLine()); }
	void PostNote(const std::string & msg, unsigned int line);

	wtast::Program* GetProgram();

	const UniqueId             FirstUniqueId;

	static const bool          EnableLexerDebug   = true;
	static const bool          EnableRuleDebug    = true;
	static const wt_u32        FirstPublicTypeId  = 0;
	static const wt_u32        FirstPrivateTypeId = WTTM_USERTYPEID >> 1;
	static const wt_u32        TypeIdRange        = WTTM_USERTYPEID >> 1;
	static const bool          LeftAssoc          = true;
	static const bool          RightAssoc         = false;
	static const wt_i32        MinPrecedence      = WTBSYM_OP_MINPREC;
	static const wt_i32        MaxPrecedence      = WTBSYM_OP_MAXPREC;
	static const std::string   GlobalBlockName;
	static const std::string   MainBlockName;
	static const std::string   CtorBlockName;
	static const std::string   DtorBlockName;
	static const std::string   StrLeftAssoc;
	static const std::string   StrRightAssoc;
	static const std::string   OperatorChars;

	friend class wtparser::WTParser;

public:
	static int  Lexer_Operator(void* lexer);
	static bool Lexer_HaveLocalType(void* lexer);
	static int  Lexer_TransformIdent(void* lexer);

public:
	static bool        IsAnyOf(char ch, const std::string & in) { return in.find_first_of(ch)!=std::string::npos; }
	static bool        IsHexDigit(char ch) { static const std::string digits = "0123456789ABCDEFabcdef"; return IsAnyOf(ch, digits); }
	static wt_u32      HexValue(char);

	std::string GetNumberCore(const std::string &);
	wt_u8       BinToI8 (const std::string &);
	wt_u8       OctToI8 (const std::string &);
	wt_u8       DecToI8 (const std::string &);
	wt_u8       HexToI8 (const std::string &);
	wt_u16      BinToI16(const std::string &);
	wt_u16      OctToI16(const std::string &);
	wt_u16      DecToI16(const std::string &);
	wt_u16      HexToI16(const std::string &);
	wt_u32      BinToI32(const std::string &);
	wt_u32      OctToI32(const std::string &);
	wt_u32      DecToI32(const std::string &);
	wt_u32      HexToI32(const std::string &);
	wt_u64      BinToI64(const std::string &);
	wt_u64      OctToI64(const std::string &);
	wt_u64      DecToI64(const std::string &);
	wt_u64      HexToI64(const std::string &);
	wt_f32      ParseF32(const std::string &);
	wt_f64      ParseF64(const std::string &);
	wt_u32      ParseChar(const char* &);
	wt_u8       strto8 (const char* str, wt_u32 base);
	wt_u16      strto16(const char* str, wt_u32 base);
	wt_u32      strto32(const char* str, wt_u32 base);
	wt_u64      strto64(const char* str, wt_u32 base);

public:
	struct Usage
	{
		enum Kind
		{
			Assign,
		};
	};

public:
	char                             ParseChar(const std::string &);
	std::string                      ParseString(const std::string &);
	std::string                      FileAsString();
	std::string                      LineAsString();
	std::string                      DateAsString();
	std::string                      TimeAsString();
	wt_u32                           GetTypeId(const std::string &);
	wt_u32                           GetSeqTypeId(wtast::Type*);
	UniqueId                         GetFuncUniqueId(wtast::FunctionSignature*);
	UniqueId                         CreateUniqueId();
	wtast::Type*                     GetStockType(wt_u32 basictype_id);
	wtast::Type*                     GetResolvedType(wt_u32 type_id);
	wtast::SeqType*                  GetSeqType(wtast::Type* type);
	wtast::Type*                     ResolveType(const std::string &);
	wtast::FunctionSignature*        ResolveFunction(const std::string & funcname, wtast::TypeInfo* typeinfo);
	wtast::Expression*               ResolveFunctionCall(wtast::Expression* funcexpr, wtast::Expression* argexpr);
	wtast::UnresolvedFunctionName*   ResolveGlobalFunctionName(const std::string & funcname);
	bool                             ExpressionHasNoType(wtast::Expression* expr);
	bool                             IsLabelDefined(const std::string & once_label);
	bool                             HaveLocalType(const std::string & str);
	bool                             IsSpecialIdent(const std::string & str);
	int                              OperatorLookup(const std::string & str);
	int                              TransformIdent(const std::string & str);
	bool                             IsBoolType(wtast::TypeInfo* typeinfo);
	bool                             IsFunctionType(wtast::TypeInfo* typeinfo);
	bool                             IsSameType(wtast::Type* lhs, wtast::Type* rhs, Usage::Kind usage=Usage::Assign);
	wtast::RecordField*              MakeRecordField(wtast::Type* type, const std::string & name);
	wtast::FunctionDeclaration*      MakeFunctionDeclaration(wtast::FunctionSignature*);
	wtast::FunctionDefinition*       MakeFunctionDefinition(wtast::FunctionSignature* funcsig, wtast::PreCondition* precond, wtast::PostCondition* postcond, wtast::ExceptionCompound* ec);
	wtast::RecordDeclaration*        MakeRecordDeclaration(const std::string & name, recordfields_t & fields, bool priv, bool lim, bool restricted);
	wtast::ExceptionDeclaration*     MakeExceptionDeclaration(const std::string & str);
	wtast::OperatorDeclaration*      MakeOperatorDeclaration(const std::string & opname, const std::string & assoc, wt_i32 prec);
	wtast::/*Alias*/Declaration*         MakeAliasDeclaration(const std::string & alias, wtast::Type* type);
	wtast::FunctionTypeDeclaration*  MakeFunctionTypeDeclaration(wtast::FunctionSignature* sig);
	wtast::FunctionPointerCall*      MakeFunctionPointerCall(wtast::FunctionType* functype, wtast::Expression* funcptraccess, wtast::Expression* argexpr);
	wtast::MakeFunctionPointer*      MakeMakeFunctionPointer(const std::string & functype_name, const std::string & funcname);
	wtast::DataType*                 MakeDataType(const std::string & name);
	wtast::DataTypeCons*             MakeDataTypeCons(const std::string & name);
	wtast::DataTypeDeclaration*      MakeDataTypeDeclaration(wtast::DataType* datatype, datatypeconses_t & conses);
	wtast::EnumType*                 MakeEnumType(const std::string & name);
	wtast::Enumerator*               MakeEnumerator(const std::string & name);
	wtast::EnumTypeDeclaration*      MakeEnumTypeDeclaration(wtast::EnumType* enumtype, enumerators_t & enumerators);
	void                             SetCurrentDataType(wtast::DataType* datatype);
	void                             SetCurrentEnumType(wtast::EnumType* enumtype);
	void                             AddDeclaration(wtast::Declaration* decl);
	wtast::LValue*                   ResolveMemberAccess(wtast::LValue* lvalue, const std::string & fieldname);
	wtast::ReturnStatement*          MakeReturnStatement(wtast::Expression* expr);
	wtast::Expression*               MakeOrElse(wtast::Expression* lhs, wtast::Expression* rhs);
	wtast::Expression*               MakeAndThen(wtast::Expression* lhs, wtast::Expression* rhs);
	wtast::Expression*               ApplyBinaryOperator(wtast::Expression* lhs, const std::string & op, wtast::Expression* rhs);
	wtast::Expression*               ApplyUnaryOperator(wtast::Expression* lhs, const std::string & op);
	void                             EnterCurrentFunction(wtast::FunctionSignature*);
	void                             LeaveCurrentFunction();
	void                             EnterEHandlerMode(bool);
	void                             LeaveEHandlerMode();
	void                             EnterFinallyMode(bool);
	void                             LeaveFinallyMode();
	void                             EnterDistinguishedBlockMode(bool);
	void                             LeaveDistinguishedBlockMode();
	int                              GetOperatorToken(bool left_assoc, wt_i32 prec);
	wtast::VariableDeclaration*      EnterForeachBlock(const std::string & varname, wtast::Expression* from, bool range_based);
	wtast::RaiseStatement*           MakeRaiseStatement(const std::string & exc, bool near);
	wtast::WhileStatement*           MakeWhileStatement(wtast::Expression* cond, wtast::Compound* body);
	wtast::DoWhileStatement*         MakeDoWhileStatement(wtast::Compound* body, wtast::Expression* cond);
	wtast::ForStatement*             MakeForStatement(wtast::Assignment* init, wtast::Expression* cond, wtast::Assignment* step, wtast::Compound* body);
	wtast::ForeachStatement*         MakeForeachStatement(wtast::VariableDeclaration* vardecl, wtast::Expression* from, wtast::Expression* to, bool inclusive, bool reverse, wtast::Expression* whil, wtast::Expression* whic, wtast::Compound* body);
	wtast::ForeachStatement*         MakeForeachStatement(wtast::VariableDeclaration* vardecl, wtast::Expression* container, bool reverse, wtast::Expression* whil, wtast::Expression* whic, wtast::Compound* body);
	wtast::FunctionArgument*         MakeFunctionArgument(wtast::Type* type, const std::string & name, bool is_mutable);
	wtast::IfStatement*              MakeIfStatement(wtast::Expression* cond, wtast::Compound* then, wtast::Statement* els);
	wtast::AssertStatement*          MakeAssertStatement(wtast::Expression* assertion);
	wtast::PreCondition*             MakePrecondition(wtast::Expression* cond);
	variabledecls_t*                 AddPostcondRets(identlist_t* rets);
	variabledecls_t*                 AddPostcondOlds(identlist_t* olds);
	wtast::PostCondition*            MakePostcondition(variabledecls_t* rets, variabledecls_t* olds, wtast::Expression* cond);
	wtast::LValue*                   ResolveVariableAccess(const std::string & varname, const std::string & scopename);
	wtast::SeqLiteral*               MakeSeqLiteral(wtast::Type* elemtype, wtast::Expression* init);
	wtast::RecordLiteral*            MakeRecordLiteral(const std::string & type_name, wtast::Expression* initexpr, fieldassigns_t * assigns);
	wtast::RecordUpdate*             MakeRecordUpdate(wtast::Expression* expr, fieldassigns_t * assigns);
	wtast::DataTypeLiteral*          MakeDataTypeLiteral(const std::string & cons_name, wtast::Expression* expr);
	wtast::DataTypeLiteral*          MakeDataTypeLiteral(const std::string & dt_name, const std::string & cons_name, wtast::Expression* expr);
	wtast::DataTypeLiteral*          MakeDataTypeLiteral(wtast::DataType* dt, wtast::DataTypeCons* cons, wtast::Expression* expr);
	wtast::SwitchStatement*          MakeSwitchStatement(wtast::Expression* cond, switchcases_t* cases, wtast::DefaultStatement* defstmt);
	wtast::AstNode::Kind             DecodeCaseKind(wtast::Type* type);
	const std::string &              DecodeCaseName(wtast::SwitchCase* sc);
	variabledecls_t*                 AddCaseVariables(const std::string & cons_name, identlist_t* ids);
	variabledecls_t*                 AddCaseVariables(const std::string & dt_name, const std::string & cons_name, identlist_t* ids);
	variabledecls_t*                 AddCaseVariables(wtast::DataType* dt, wtast::DataTypeCons* cons, identlist_t* ids);
	wtast::DataTypeCaseStatement*    MakeDataTypeCaseStatement(const std::string & cons_name, variabledecls_t* vars, wtast::Compound* comp);
	wtast::DataTypeCaseStatement*    MakeDataTypeCaseStatement(const std::string & dt_name, const std::string & cons_name, variabledecls_t* vars, wtast::Compound* comp);
	wtast::DataTypeCaseStatement*    MakeDataTypeCaseStatement(wtast::DataType* dt, wtast::DataTypeCons* cons, variabledecls_t* vars, wtast::Compound* comp);
	wtast::SwitchCase*               MakeSpecialCaseStatement(const std::string & special_name, wtast::Compound* comp);
	wtast::SwitchCase*               MakeSpecialCaseStatement(const std::string & type_name, const std::string & special_name, wtast::Compound* comp);
	wtast::DefaultStatement*         MakeDefaultStatement(wtast::Compound* comp);
	wtast::Expression*               MakeSpecialLiteral(const std::string & special_name);
	wtast::Expression*               MakeSpecialLiteral(const std::string & type_name, const std::string & special_name);
	wt_u32                           GetNextContextSpecificTypeId();
	wt_u32                           ResolveTypeIdAliasing(wt_u32 type_id);
	bool                             ExpressionCanBeArgument(wtast::FunctionSignature* sig, wtast::Expression* expr);
	bool                             ExpressionIsMutable(wtast::Expression* expr);
	bool                             ExpressionIsCopyable(wtast::Expression* expr);
	wtast::Expression*               ValidateExpression(wtast::Expression* expr);
	void                             ValidateAssignment(wtast::Assignment* as);
	bool                             ValidateInitialValue(wtast::VariableDeclaration* vardecl, wtast::Expression* init);
	bool                             ValidateUninitializedVariable(wtast::VariableDeclaration* vardecl);
	wtast::AstNode::Variable::Class  GetCurrentVariableClass() const { return mCurrentVariableClass; }
	void                             SetCurrentVariableClass(wtast::AstNode::Variable::Class cls) { mCurrentVariableClass = cls; }
	std::string                      ValidateFunctionName(const std::string & funcname, bool raw_string=true);
	void                             RegisterFunctionSignature(wtast::FunctionSignature* funcsig);
	void                             HideVariables(wtast::VariableDeclaration* decl);
	bool                             FunctionHasPfnInArgs(wtast::FunctionSignature* sig);
	void                             PushSwitchExpression(wtast::Expression* expr);
	void                             PopSwitchExpression();
	bool                             CanBeIndexed(wtast::TypeInfo* info, wtast::TypeInfo* eti);
	bool                             ExpressionHasUniqueId(wtast::Expression* expr, const std::set<UniqueId> & uids);

private:
	struct NameTypePair
	{
		NameTypePair(const std::string & n, wtast::TypeInfo* t)
			: name(n)
			, type(t)
		{
			// ok
		}

		friend bool operator<(const NameTypePair & lhs, const NameTypePair & rhs)
		{
			return less(lhs, rhs);
		}

		std::string      name;
		wtast::TypeInfo* type;

		static bool less(const NameTypePair &, const NameTypePair &);
	};

	struct RawTypes
	{
		typedef std::vector<wt_u32> types_t;
		typedef types_t::const_iterator const_iterator;
		types_t types;

		const_iterator begin() const { return types.begin(); }
		const_iterator end()   const { return types.end();   }
		std::size_t    size()  const { return types.size();  }

		RawTypes() { }
		RawTypes(wt_u32 t) { types.push_back(t); }
		RawTypes(const types_t & ts) : types(ts.begin(), ts.end()) { }

		RawTypes & Add(wt_u32 t)
		{
			types.push_back(t);
			return *this;
		}
	};

	struct OperatorTokenPair
	{
		OperatorTokenPair() : op(0), tok(0) { }
		OperatorTokenPair(wtast::OperatorDeclaration* op, int tok) : op(op), tok(tok) { }

		wtast::OperatorDeclaration* op;
		int                         tok;
	};

	bool                      _AddOperator(const std::string & opname, bool left_assoc, wt_i32 prec);
	bool                      _AddFunction(const std::string & name, const RawTypes & args, const RawTypes & rets, bool pure, bool builtin, bool generated, UniqueId unique_id);
	bool                      _AddBuiltin(const std::string & name, const RawTypes & args, const RawTypes & rets, bool pure, UniqueId unique_id);
	bool                      _AddGenerated(const std::string & name, const RawTypes & args, const RawTypes & rets, bool pure, UniqueId unique_id);
	bool                      _AddSequenceOperations(wtast::Type* base_type, wt_u32 seq_typeid);
	bool                      _AddEnumOperations(wtast::EnumType* enum_type, wt_u32 enum_typeid);
	wtast::FunctionSignature* _GetCurrentFunction();
	bool                      _GetEHandlerMode();
	bool                      _GetFinallyMode();
	bool                      _GetDistinguishedBlockMode();

	typedef wtast::AstNode::Variable::Class varclass_t;

	typedef std::map<NameTypePair, wtast::FunctionSignature*> nametype_funcsig_t;
	typedef std::map<UniqueId, wtast::FunctionSignature*> map_uniqueid_funcsig_t;
	typedef std::vector<wtast::FunctionSignature*> vec_funcsig_t;
	typedef std::map<std::string, vec_funcsig_t> name_funcsig_t;
	typedef std::vector<wtast::AstNode*> allocatednodes_t;
	typedef std::map<std::string, wtast::Type*> map_name_type_t;
	typedef std::map<wt_u32, wtast::Type*> map_u32_type_t;
	typedef std::map<wt_u32, wt_u32> map_u32_u32_t;
	typedef std::map<std::string, OperatorTokenPair> map_name_optok_t;
	typedef std::map<std::string, UniqueId> map_name_uniqueid_t;
	typedef std::map<std::string, std::set<wtast::DataTypeCons*> > map_name_datatypeconses_t;
	typedef std::map<std::string, std::set<wtast::Enumerator*> > map_name_enumerators_t;
	typedef std::map<wtast::DataTypeCons*, wtast::DataType*> map_cons_datatype_t;
	typedef std::map<wtast::Enumerator*, wtast::EnumType*> map_enumerator_enumtype_t;
	typedef std::vector<wtast::Expression*> vec_expr_t;
	typedef std::vector<bool> vec_bool_t;

public:
	Debugger<EnableLexerDebug> mLexerDebugger;
	Debugger<EnableRuleDebug>  mRuleDebugger;

private:
	wtparser::SymTab           mSymTab;
	wtparser::SourceManager    mSrcMan;
	wtparser::ProblemHandler   mProbHan;
	wtast::Program*            mProgram;
	wtast::DeclareBlock*       mDeclEntity;
	std::set<std::string>      mOnceLabels;
	vec_funcsig_t              mCurrentFunctionStack;
	vec_bool_t                 mEHandlerModeStack;
	vec_bool_t                 mFinallyModeStack;
	vec_bool_t                 mDistinguishedBlockModeStack;
	bool                       mSpecialBlockMode;
	std::string                mCompileDate;
	std::string                mCompileTime;
	nametype_funcsig_t         mNametypeFuncsig;
	map_uniqueid_funcsig_t     mUniqueidFuncsig;
	name_funcsig_t             mNameFuncsig;
	allocatednodes_t           mAllocatedAstNodes;
	unsigned int               mNextUniqueId;
	wt_u32                     mNextPublicTypeId;
	wt_u32                     mNextPrivateTypeId;
	varclass_t                 mCurrentVariableClass;
	wtast::DataType*           mCurrentDataType;
	wtast::EnumType*           mCurrentEnumType;
	std::set<std::string>      mLocalTypenames;
	map_name_type_t            mTypenameToType;
	map_u32_type_t             mTypeidToType;
	map_u32_u32_t              mAliasToId;
	map_u32_u32_t              mAutoAliases;
	map_name_optok_t           mOpnameToOp;
	map_name_uniqueid_t        mExceptionToId;
	map_name_datatypeconses_t  mNameToDataTypeConses;
	map_cons_datatype_t        mConsToDataType;
	map_name_enumerators_t     mNameToEnumerators;
	map_enumerator_enumtype_t  mEnumeratorToEnumType;
	vec_expr_t                 mSwitchExpressionStack;
};

} // wtparser

#endif // WTPARSER_PARSERHELPER_H
